global function OnWeaponPrimaryAttack_shifter
global function MpAbilityShifterWeapon_Init
global function MpAbilityShifterWeapon_OnWeaponTossPrep
global function AbilityShifter_ApplyInProgressStimIfNeeded

const SHIFTER_WARMUP_TIME = 0.0
const SHIFTER_WARMUP_TIME_FAST = 0.0

const string PHASEEXIT_IMPACT_TABLE_PROJECTILE	= "default"
const string PHASEEXIT_IMPACT_TABLE_TRACE		= "superSpectre_groundSlam_impact"

const float PMMOD_ENDLESS_STRENGTH = 0.8

struct
{
	int phaseExitExplodeImpactTable
} file;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#if SERVER
//per player data
global table<string, int> playercyclestagetable
global table<string,float> playerchargedtimetable
global table<string, vector> playerorigpostable //stores the original position after a player used his boost
global table<string, array<vector> > playerinprogresspos //stores an array of location that are tracked after the player used his boost
global table<string, entity> playerrewindindicatortable //stores that players rewind indicator


global function FindNearestSafePos
#endif
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void function MpAbilityShifterWeapon_Init()
{
	// "exp_rocket_archer"
	// "exp_xlarge"
	// "exp_arc_ball"
	file.phaseExitExplodeImpactTable = PrecacheImpactEffectTable( PHASEEXIT_IMPACT_TABLE_PROJECTILE )
	PrecacheImpactEffectTable( PHASEEXIT_IMPACT_TABLE_TRACE )
	PrecacheParticleSystem($"P_elec_arc_loop_LG_1")

	#if SERVER
	AddCallback_OnClientConnected(perplayerdatainit)
	AddCallback_OnPlayerRespawned(perplayerdatainit)
	PrecacheModel( $"models/fx/core_energy.mdl")
	#endif
}

/////////////////////////////////////////////////custom init////////////////////////////////////////////////////
#if SERVER
void function perplayerdatainit(entity player) {
	playercyclestagetable[player.GetPlayerName()] <- 0
	playerchargedtimetable[player.GetPlayerName()] <- 0

	playerorigpostable[player.GetPlayerName()] <- player.GetOrigin()
	playerinprogresspos[player.GetPlayerName()] <- []

}
#endif
////////////////////////////////////////////////////////////////////////////////////////////////////////////////


void function MpAbilityShifterWeapon_OnWeaponTossPrep( entity weapon, WeaponTossPrepParams prepParams )
{
	entity weaponOwner = weapon.GetWeaponOwner()
	int pmLevel = GetPVEAbilityLevel( weapon )
	if ( (pmLevel >= 2) && IsValid( weaponOwner ) && weaponOwner.IsPhaseShifted() )
		weapon.SetScriptTime0( Time() )
	else
		weapon.SetScriptTime0( 0.0 )
}

int function GetPVEAbilityLevel( entity weapon )
{
	if ( weapon.HasMod( "pm2" ) )
		return 2
	if ( weapon.HasMod( "pm1" ) )
		return 1
	if ( weapon.HasMod( "pm0" ) )
		return 0

	return -1
}



//////////////////////////////////////////////custom act func///////////////////////////////////////////////////
var function OnWeaponPrimaryAttack_shifter( entity weapon, WeaponPrimaryAttackParams attackParams )
{
	#if SERVER
	entity player = weapon.GetOwner()

	//tick up cyclestage
	playercyclestagetable[player.GetPlayerName()] <- playercyclestagetable[player.GetPlayerName()] + 1


	if (playercyclestagetable[player.GetPlayerName()] == 1) {
		if (GetConVarBool("phr_enable_chargeup")) { //normal
			thread startchargetimer(player)
			thread startchargetimeout(player,weapon)
			weapon.SetWeaponPrimaryClipCount(400)
		} else { //skip to boost and set charge up to 100%
			playercyclestagetable[player.GetPlayerName()] = 2
			playerchargedtimetable[player.GetPlayerName()] = 1
		}

	}



	if (playercyclestagetable[player.GetPlayerName()] == 2) {
		boostplayer(player)
		weapon.SetWeaponPrimaryClipCount(200)

		if (GetConVarBool("phr_enable_phase_rewind") == true) {
			thread startrewindtimeout(player, weapon)
			try{
				entity indicatorent = createindicator(playerorigpostable[player.GetPlayerName()])
				playerrewindindicatortable[player.GetPlayerName()] <- indicatorent
			}catch(ex){print(ex)} //might not be a vector present

		} else { //end cycle here
			playercyclestagetable[player.GetPlayerName()] <- 0
			thread startcooldown(player,weapon)
			weapon.SetWeaponPrimaryClipCount(0)
		}

	}

	if (playercyclestagetable[player.GetPlayerName()] == 3) {
		thread rewindplayer(player)
		weapon.SetWeaponPrimaryClipCount(0)
	}


	//reset when 3rd stage was done
	if (playercyclestagetable[player.GetPlayerName()] == 3) {
		playercyclestagetable[player.GetPlayerName()] <- 0
		thread startcooldown(player,weapon)
	}
	#endif
	return 0 //how much "ammo" is used , 0 because its getting set manually
}
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////custom functions////////////////////////////////////////////////////////
#if SERVER
void function startchargetimer(entity player) {
	thread startchargeindicator(player)
	playerchargedtimetable[player.GetPlayerName()] <- 0
	while (playercyclestagetable[player.GetPlayerName()] == 1 && playerchargedtimetable[player.GetPlayerName()] < 1.0 && IsAlive(player)) {

		string bar = ""
		for (int i = int(playerchargedtimetable[player.GetPlayerName()] *10);i>0;i--) {
			bar = bar + "|||"
		}

		playerchargedtimetable[player.GetPlayerName()] = playerchargedtimetable[player.GetPlayerName()] + 0.1

		SendHudMessage(player,"Charging: \n"+bar,-1,0.65,255,255,255,1,0,0.35,0) //player //text //x //y //r //g //b  /a  //fade-in s //lenght s //fade-out s

		wait 0.3
	}
	if (playercyclestagetable[player.GetPlayerName()] == 1) {SendHudMessage(player,"Fully Charged",-1,0.65,255,255,255,1,0,0.5,0.8)}
}

void function startchargetimeout(entity player,entity weapon) {
	wait 8 //3 seconds of charge up + 5 sec hold window
	if (IsValid(player) && IsValid (weapon) && playercyclestagetable[player.GetPlayerName()] == 1) {
		SendHudMessage(player,"Charge Depleted",-1,0.65,255,255,255,1,0,0.5,0.8)
		weapon.SetWeaponPrimaryClipCount(600)
		playercyclestagetable[player.GetPlayerName()] = 0
	}

}

void function startchargeindicator(entity player) { //visual effect
	while (IsValid(player) && playercyclestagetable[player.GetPlayerName()] == 1 && IsAlive(player)) {
		PlayFX($"P_elec_arc_loop_LG_1", player.GetOrigin()+/*player.GetViewVector()*20+*/<0,0,40>)
		WaitFrame()
	}
}

//////////////////

void function startanimationposrecording(entity player) {
	playerinprogresspos[player.GetPlayerName()] <- []
	array <vector> positionarray = []

	while (IsValid(player) && playercyclestagetable[player.GetPlayerName()] == 2) {
		//should probably implement a check here so a location thats to close to the one before isnt stored
		try {
			positionarray.push(player.GetOrigin())
			wait 0.35
			playerinprogresspos[player.GetPlayerName()] <- positionarray
		}catch(ex) {print(ex)}
	}

}

//////////////////

void function startrewindtimeout(entity player, entity weapon) {
	//wait 5
	try {
		string startbar = "||||||||||||||||||||||||||||||"
		for (int i=0;i<30;i++) {
			if (playercyclestagetable[player.GetPlayerName()] != 0) {
				string bar = startbar.slice(i,29)
				SendHudMessage(player,"Rewind Window: \n"+bar,-1,0.65,255,255,255,1,0,0.2,0)
				wait 0.166666
			}
		}

		if (playercyclestagetable[player.GetPlayerName()] != 0) {
			if (IsValid(player) && IsValid(weapon)) {
				playercyclestagetable[player.GetPlayerName()] <- 0
				weapon.SetWeaponPrimaryClipCount(0)
				thread startcooldown(player,weapon)
			}
		}
	}catch(ex) {print(ex)}
}

void function startcooldown (entity player,entity weapon) {
	int cooldownint = GetConVarInt("phr_cooldown")
	wait cooldownint
	if (IsValid(player) && IsValid(weapon)) {
		if (weapon.GetWeaponPrimaryClipCount() == 0) {
			weapon.SetWeaponPrimaryClipCount(600)
			playercyclestagetable[player.GetPlayerName()] <- 0
		}
	}

}

//////////////////

void function boostplayer (entity player) {

	float velmultiplier = GetConVarFloat("phr_boost_strength")

	playerorigpostable[player.GetPlayerName()] <- player.GetOrigin()
	thread startanimationposrecording(player)

	PhaseShift(player, 0, 0.3 )
	vector viewvector = player.GetViewVector() * 1000 * playerchargedtimetable[player.GetPlayerName()] * velmultiplier
	player.SetVelocity(viewvector)
}

void function rewindplayer (entity player) {
	try {
	string playername = player.GetPlayerName()
	vector origpos = playerorigpostable[player.GetPlayerName()]

	if(GetConVarBool("phr_enable_rewind_animation") == true) {
		PhaseShift(player, 0, 0.1 * playerinprogresspos[player.GetPlayerName()].len()) //phaseshift for as long as the mover is going to take for all the in progress pos's

		entity mover = CreateOwnedScriptMover (player)
		player.SetParent(mover)

		for (int i = playerinprogresspos[player.GetPlayerName()].len() - 1; i >= 0; i--) { //loop through all the tracked position backwards without checking for clipping

			mover.NonPhysicsMoveTo (playerinprogresspos[player.GetPlayerName()][i], 0.1,0,0.0)
			WaitFrame() //arbitrarily low number , dk didnt get it to go faster

		}

		playerinprogresspos[player.GetPlayerName()] <- []
		player.SetVelocity(<0,0,0>)
		player.ClearParent()
		mover.Destroy()

		player.SetOrigin(FindNearestSafePos(player, origpos,1)) //after animation is done , tp to actual original position
	} else {
		float movetime = 0.5
		PhaseShift(player, 0, movetime)

		entity mover = CreateOwnedScriptMover (player)
		player.SetParent(mover)

		mover.NonPhysicsMoveTo (FindNearestSafePos(player, origpos,1), movetime,0,0)
		wait movetime

		player.ClearParent()
		player.SetVelocity(<0,0,0>)
		mover.Destroy()

		player.SetOrigin(FindNearestSafePos(player, origpos,1)) //after mover and phase is done , just tp anyway
	}

	deleteindicator(playerrewindindicatortable[player.GetPlayerName()])

	}catch(ex) {print(ex)} //might disconnect when moving
}

////////////////////////////

entity function createindicator (vector location) {

	entity ind = CreateEntity( "prop_dynamic" )
	ind.SetValueForModelKey( $"models/fx/core_energy.mdl" )
	ind.SetOrigin( <location.x,location.y,location.z+50>)
	DispatchSpawn( ind )

	thread indicatortimeout(ind)

	return ind


}

void function indicatortimeout(entity indicator) {
	wait 7 //some number thats a bit larger than rewind window
	if (IsValid(indicator)) {deleteindicator(indicator)}
}

void function deleteindicator (entity indicator) {

	if (IsValid(indicator)) {
		indicator.Destroy()
	} else (print("Phase Error:"+"Deletion Object wasnt valid"))

}

//////////////////////////////////////////////utils///////////////////////////////////////////////////////////
vector function FindNearestSafePos(entity player, vector targetPos, int severity)
{
    if(severity >= 600)
    {
        print("Phase error:could not find a save position for player")
       return < player.GetOrigin().x,player.GetOrigin().y,player.GetOrigin().z+1500 >
    }


	//check if adding/removing severity  from the x,y or z coords will stop the player from clipping

	//adding / removing y
    if( !PlayerPosInSolid( player, < targetPos.x, targetPos.y+severity, targetPos.z > ) )
        return < targetPos.x, targetPos.y+severity, targetPos.z >

    if( !PlayerPosInSolid( player, < targetPos.x, targetPos.y-severity, targetPos.z > ) )
        return < targetPos.x, targetPos.y-severity, targetPos.z >

	//adding / removing x
    if( !PlayerPosInSolid( player, < targetPos.x+severity, targetPos.y, targetPos.z > ) )
        return < targetPos.x+severity, targetPos.y, targetPos.z >

    if( !PlayerPosInSolid( player, < targetPos.x-severity, targetPos.y, targetPos.z > ) )
        return < targetPos.x-severity, targetPos.y, targetPos.z >

	//adding / removing z(height)
	if( !PlayerPosInSolid( player, < targetPos.x, targetPos.y, targetPos.z +severity> ) )
        return < targetPos.x, targetPos.y, targetPos.z +severity>

    if( !PlayerPosInSolid( player, < targetPos.x, targetPos.y, targetPos.z -severity > ) )
        return < targetPos.x, targetPos.y, targetPos.z -severity>


    return FindNearestSafePos( player, targetPos, severity+5 ) //severity goes up by 5 each time adding severity doesnt stop the player from clipping
}

bool function PlayerPosInSolid( entity player, vector targetPos ) //checks if a the player would be clipping if he were in targetpos
{
    int solidMask = TRACE_MASK_PLAYERSOLID
    vector mins
    vector maxs
    int collisionGroup = TRACE_COLLISION_GROUP_PLAYER
    array<entity> ignoreEnts = []
    ignoreEnts.append( player ) //in case we want to check player's current pos
    TraceResults result

    mins = player.GetPlayerMins()
    maxs = player.GetPlayerMaxs()
    result = TraceHull( targetPos, targetPos + Vector( 0, 0, 1), mins, maxs, ignoreEnts, solidMask, collisionGroup )
    if ( result.startSolid )
        return true

    return false

}


#endif
////////////////////////////////////////////////////////////////////////////////////////////////////////////////


void function ApplyInProgressStimIfNeededThread( entity player )
{
	wait 0.5  // timed to kick in when the notify appears on player's screen

	if ( !IsAlive( player ) )
		return
	if ( !player.IsPhaseShifted() )
		return

	EndlessStimBegin( player, PMMOD_ENDLESS_STRENGTH )
}

void function AbilityShifter_ApplyInProgressStimIfNeeded( entity player )
{
	thread ApplyInProgressStimIfNeededThread( player )
}

void function DoPhaseExitExplosion( entity player, entity phaseWeapon )
{
#if CLIENT
	if ( !phaseWeapon.ShouldPredictProjectiles() )
		return
#endif //

	player.PhaseShiftCancel()

	vector origin = player.GetWorldSpaceCenter() + player.GetForwardVector() * 16.0

	//DebugDrawLine( player.GetWorldSpaceCenter(), origin, 255, 0, 0, true, 5.0 )

	int damageType = (DF_RAGDOLL | DF_EXPLOSION | DF_ELECTRICAL)
	entity nade = phaseWeapon.FireWeaponGrenade( origin, <0,0,1>, <0,0,0>, 0.01, damageType, damageType, true, true, true )
	if ( !nade )
		return

	player.PhaseShiftBegin( 0, 1.0 )

	nade.SetImpactEffectTable( file.phaseExitExplodeImpactTable )
	nade.GrenadeExplode( <0,0,0> )

#if SERVER
	PlayImpactFXTable( player.GetOrigin(), player, PHASEEXIT_IMPACT_TABLE_TRACE, SF_ENVEXPLOSION_INCLUDE_ENTITIES )
#endif //
}



//note to self: FindNearestSafePos, PlayerPosInSolid and the ground/roof checking was done by VoyageDB#5167 on discord